/***************************************************************************
 *	evolmixer.c: Żҥܥ塼ࡦߥ塼ȥ쥸
 *	
 *			Programed by Nobuyuki Maruichi
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>

/* #include <linux/fcntl.h> */
#include <linux/soundcard.h>

#ifdef HAVE_LINUX_TOWNSVOL_H
# include <linux/townsvol.h>
#endif /*HAVE_LINUX_TOWNSVOL_H*/

#define EVOL_VERSION "evolmixer-1.0.1"

int string2level(const char *target, int ovol);

#ifdef TOWNS
/***************************************************************************
 *	TownsStandardVolumes:  evol Ǥλ˽򤹤뤿Υǡ
 *	С
 *		shortName:	Υͥ̾
 *		device:		USS  ioctl Υܥ塼оݤ򼨤
 *		newLevel:	ܥ塼ͤѹˡʴؿؤΥݥ󥿡
 *		
 ***************************************************************************/
typedef struct {
	char *shortName;
	int device;
	int (*newLevel)(int volume, int new);
} TownsStandardVolumes;

int monoLevel(int volume, int left);
int leftLevel(int volume, int left);
int rightLevel(int volume, int right);

TownsStandardVolumes TownsEVolumes[] = {
	{ "ll",   SOUND_MIXER_LINE, leftLevel},
	{ "lr",   SOUND_MIXER_LINE, rightLevel},
	{ "al", SOUND_MIXER_ALTPCM, leftLevel},
	{ "ar", SOUND_MIXER_ALTPCM, rightLevel},
	{ "cl",     SOUND_MIXER_CD, leftLevel},
	{ "cr",     SOUND_MIXER_CD, rightLevel},
	{ "mc",    SOUND_MIXER_MIC, monoLevel},
	{ "md",  SOUND_MIXER_LINE1, monoLevel},
	{ NULL, 0, NULL}
};

int OldDevicesInfo = SOUND_MASK_LINE | SOUND_MASK_ALTPCM | 
						SOUND_MASK_CD | SOUND_MASK_MIC | SOUND_MASK_LINE1;
#endif /*TOWNS*/

#define NumberOfArray(array) (sizeof(array)/sizeof(array[0]))

/***************************************************************************
 *	: 
 *		DeviceInfoProc: ܥ塼ǥХɽؿ
 *		: 
 *		
 ***************************************************************************/
typedef int (*DeviceInfoProc)(int mixer, const char *target, 
					int devices, int streo, int counter, int co);
typedef struct {
	int limit;
	const char **names;
} DeviceNamesInfo;

/***************************************************************************
 *	Хѿ: 
 *		Columns: ɽʴĶѿ
 *		DeviceNames: إå줿ǥХ̾
 *		
 ***************************************************************************/
int Columns = 80;
const char *SoundDeviceNames[]   = SOUND_DEVICE_NAMES;
const char *SoundDeviceLabels[]  = SOUND_DEVICE_LABELS;
#ifdef TOWNS
const char *TSoundDeviceNames[]  = TSOUND_DEVICE_NAMES;
const char *TSoundDeviceLabels[] = TSOUND_DEVICE_LABELS;
#endif /*TOWNS*/

#ifdef TOWNS
DeviceNamesInfo DevicesNames[] = {
	{NumberOfArray(TSoundDeviceNames),  TSoundDeviceNames}, 
	{NumberOfArray(TSoundDeviceLabels), TSoundDeviceLabels}, 
	{NumberOfArray(SoundDeviceNames),   SoundDeviceNames}, 
	{NumberOfArray(SoundDeviceLabels),  SoundDeviceLabels}, 
	{0, NULL} 
};
#else
DeviceNamesInfo DevicesNames[] = {
	{NumberOfArray(SoundDeviceNames),  SoundDeviceNames}, 
	{NumberOfArray(SoundDeviceLabels), SoundDeviceLabels}, 
	{0, NULL} 
};
#endif /*TOWNS*/

DeviceNamesInfo *CurrentDevicesNames = DevicesNames;

#ifdef TOWNS
/***************************************************************************
 *	monoLevel: Υ롦ܥ塼ѹ뤿ͤ
 *	
 *		int monoLevel(int volume, int mono);
 *	
 *		int volume;		Υܥ塼
 *		int mono;		ܥ塼
 *	֤
 *		ܥ塼ꤹ뤿
 *	
 *		
 ***************************************************************************/
int monoLevel(int volume, int mono)
{
	volume = (mono & 0x000000ff);
	volume |= (volume << 8);
	return(volume);
}

/***************************************************************************
 *	leftLevel: ƥ쥪ܥ塼κܥ塼ѹ뤿ͤ
 *	
 *		int leftLevel(int volume, int left);
 *	
 *		int volume;		Υܥ塼
 *		int left;		ܥ塼͡L
 *	֤
 *		ܥ塼ꤹ뤿
 *	
 *		
 ***************************************************************************/
int leftLevel(int volume, int left)
{
	volume &= ~0x000000ff;
	volume |= (left & 0x000000ff);
	return(volume);
}

/***************************************************************************
 *	rightLevel: ƥ쥪ܥ塼αܥ塼ѹ뤿ͤ
 *	
 *		int rightLevel(int volume, int right);
 *	
 *		int volume;     Υܥ塼
 *		int right;		ܥ塼͡R
 *	֤
 *		ܥ塼ꤹ뤿
 *	
 *		
 ***************************************************************************/
int rightLevel(int volume, int right)
{
	volume &= ~0x0000ff00;
	volume |= (right << 8) & 0x0000ff00;
	return(volume);
}

/***************************************************************************
 *	newLevel: ܥ塼٥ꤹ
 *	
 *		void newLevel(int mixer, TownsStandardVolumes *evol, int level);
 *	
 *		int mixer;						/dev/mixer Υեǥץ
 *		TownsStandardVolumes *evol;		ꤹܥ塼ǥХ
 *		int level;						ꤹܥ塼
 *	
 *		
 ***************************************************************************/
void newLevel(int mixer, TownsStandardVolumes *evol, int level)
{
	int volume;
	int devices;

	if(ioctl(mixer, SOUND_MIXER_READ_DEVMASK, &devices) >= 0) {
		if(devices & (1 << evol->device)) {
			ioctl(mixer, MIXER_READ(evol->device), &volume);
			volume = evol->newLevel(volume, level);
			ioctl(mixer, MIXER_WRITE(evol->device), &volume);
		}
	}
}

/***************************************************************************
 *	townsStandardTarget: TOWNS ɸΥܥ塼Υͥ
 *	
 *		int townsStandardTarget(int mixer, int *argc_ptr, char ***argv_ptr);
 *	
 *		int mixer;			
 *		int *argc_ptr;		
 *		char ***argv_ptr;	
 *	֤
 *		0:		ܥ塼ϹԤʤʤä
 *		0ʳ:	ꤷǥХΥӥåȥޥｪλ
 *	
 *		
 ***************************************************************************/
int townsStandardTarget(int mixer, int *argc_ptr, char ***argv_ptr, 
			int (*print)(int mixer, const char *target, 
					int devices, int streo, int counter, int co))
{
	int level, olevel;
	int counter;
	int devices;
	int streoDevices;
	int result = 0;

	if(ioctl(mixer, SOUND_MIXER_READ_DEVMASK, &devices) < 0 ||
			ioctl(mixer, SOUND_MIXER_READ_STEREODEVS, &streoDevices) < 0) {
		fprintf(stderr, "Can't use /dev/mixer\n");
		return(result);
	}
	for(counter = 0;TownsEVolumes[counter].shortName != NULL;++counter) {
		if(*argc_ptr > 1 && 
				strcmp(TownsEVolumes[counter].shortName, **argv_ptr) == 0) {
			ioctl(mixer, MIXER_READ(TownsEVolumes[counter].device), &olevel);
			olevel = olevel ^ TownsEVolumes[counter].newLevel(olevel, 0);
			olevel = ((olevel > 0x00ff)?(olevel >> 8):(olevel));
			level = string2level((const char*)(*(*argv_ptr + 1)), olevel);
			if(level != EOF) {
				newLevel(mixer, &TownsEVolumes[counter], level);
			/*
			 *	if(print != NULL) {
			 *		(void)print(mixer, NULL, devices, streoDevices, 
			 *						TownsEVolumes[counter].device, 0);
			 *	}
			 */
				--*argc_ptr;
				++*argv_ptr;
				result |= (0x0001 << TownsEVolumes[counter].device);
			}
		}
	}
	return(result);
}
#endif /*TOWNS*/

/***************************************************************************
 *	names2target: ܥ塼ǥХ̾鼱̻Ҥ
 *	
 *		int names2target(const char *target, int devices);
 *	
 *		const char *target;	ܥ塼ǥХ̾
 *		int devices;		Ѳǽʥܥ塼ǥХΥӥåȥޥ
 *	֤
 *		EOF:	ܥ塼ǥХȽǤʤä
 *		0ʾ:	ܥ塼ǥХȽǤ
 *	
 *		
 ***************************************************************************/
int names2target(const char *target, int devices, DeviceNamesInfo *names)
{
	int counter, bitCounter;
	char name_buffer[80];

	for(counter = 0, bitCounter = 0x001;
			counter < SOUND_MIXER_NRDEVICES && counter < names->limit;
										++counter, bitCounter <<= 1) {
		sscanf(names->names[counter], "%s", name_buffer);
		if((devices & bitCounter) != 0 && 
						strcmp(name_buffer, target) == 0) {
			return(counter);
		}
	}
	return(EOF);
}

/***************************************************************************
 *	name2target: ܥ塼ǥХ̾鼱̻Ҥ
 *	
 *		int name2target(const char *target, int devices);
 *	
 *		const char *target;	ܥ塼ǥХ̾
 *		int devices;		Ѳǽʥܥ塼ǥХΥӥåȥޥ
 *	֤
 *		EOF:	ܥ塼ǥХȽǤʤä
 *		0ʾ:	ܥ塼ǥХȽǤ
 *	
 *		
 ***************************************************************************/
int name2target(const char *target, int devices)
{
	int counter;
	int status = EOF;

	for(counter = 0;status == EOF && 
				DevicesNames[counter].names != NULL;++counter) {
		status = names2target(target, devices, &DevicesNames[counter]);
	}
	return(status);
}

/***************************************************************************
 *	devicePrint: ġΥǥХ̾ɽ
 *	
 *		int devicePrint(int mixer, const char *target, 
 *						int devices, int streo, int counter, int co);
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *target;		ܥ塼ǥХ̾
 *		int devices;			Ѳǽʥܥ塼Υӥåȥޥ
 *		int streo;				ƥ쥪ܥ塼Υӥåȥޥ
 *		int counter;			ܥ塼ǥХֹ
 *		int co;					ߤɽ
 *	֤
 *		ߤɽ
 *	
 *		
 ***************************************************************************/
int devicePrint(int mixer, const char *target, 
						int devices, int streo, int counter, int co)
{
	int length;
	int  bitCounter;
	char stringBuffer[20];

	bitCounter = (0x0001 << counter);

	if(devices & bitCounter) {
		if(target == NULL || name2target(target, bitCounter) == counter) {
			length = sprintf(stringBuffer, "%-8s", 
								CurrentDevicesNames->names[counter]);
			if(co + length >= Columns) {
				co = 0;
				printf("\n");
			}
			co += printf("%-8s", CurrentDevicesNames->names[counter]);
		}
	}
	return(co);
}

#ifdef TOWNS
/***************************************************************************
 *	gainForTowns: TOWNS ȼθΨɽԤʤ
 *	
 *		void gainForTowns(char *string, int volume);
 *	
 *		char *string;	ʸХåե
 *		int volume;		ܥ塼٥0-65
 *	
 *		
 ***************************************************************************/
void gainForTowns(char *string, int volume)
{
	if(volume > 0) {
		sprintf(string, "%4.1fdB", (volume - 65) / 2.0);
	} else {
		sprintf(string, "  off");
	}
}

/***************************************************************************
 *	deviceInfoTowns: ġΥǥХǥ٥ɽTOWNS Τͭ
 *	
 *		int deviceInfoTowns(int mixer, const char *target, 
 *						int devices, int streo, int counter, int co);
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *target;		ܥ塼ǥХ̾
 *		int devices;			Ѳǽʥܥ塼Υӥåȥޥ
 *		int streo;				ƥ쥪ܥ塼Υӥåȥޥ
 *		int counter;			ܥ塼ǥХֹ
 *		int co;					ߤɽ
 *	֤
 *		ߤɽʾˣˤʤ
 *	
 *		
 ***************************************************************************/
int deviceInfoTowns(int mixer, const char *target, 
						int devices, int streo, int counter, int co)
{
	int status;
	int  bitCounter;
	int level, left, right;
	int muteOnly = TOWNS_MUTE_DEVICES;
	char leftString[40];
	char rightString[40];
	const char *flag[] = {"off", "on"};

	bitCounter = (0x0001 << counter);

	if(target == NULL || name2target(target, bitCounter) == counter) {
#ifdef	OLD
	if(target == NULL || 
				strcmp(CurrentDevicesNames->names[counter], target) == 0) {
#endif	/*OLD*/
		switch(((devices & bitCounter) >> counter) +
				((streo & bitCounter) >> counter)) {
		default:
			break;
		case 1:	/* Mono */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			level &= 0x000000ff;
			if(muteOnly & bitCounter || level == 0) {
				printf("%6s  monoral  %s\n", 
					CurrentDevicesNames->names[counter], flag[level != 0]);
			} else {
				gainForTowns( leftString, USStoTowns66(level));
				printf("%6s  monoral%7s\n", 
					CurrentDevicesNames->names[counter], leftString);
			}
			break;
		case 2:	/* Stereo */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			left  = level & 0x000000ff;
			right = (level & 0x0000ff00) >> 8;
			gainForTowns( leftString, USStoTowns66( left));
			gainForTowns(rightString, USStoTowns66(right));
			printf("%6s  stereo %7s(L) %7s(R)\n", 
				CurrentDevicesNames->names[counter], leftString, rightString);
			break;
		}
	}
	return(0);
}

/***************************************************************************
 *	oldLevel:  evol ߴβɽԤʤ
 *	
 *		void oldLevel(char *string, int volume);
 *	
 *		char *string;	ʸХåե
 *		int volume;		ܥ塼٥0-65
 *	
 *		
 ***************************************************************************/
void oldLevel(char *string, int volume)
{
	switch(volume) {
	case MINIMUM_LEVEL_MUTE:
		sprintf(string, "off");
		break;
	case MINIMUM_LEVEL_1:
		sprintf(string, "min");
		break;
	case MAXIMUM_LEVEL:
		sprintf(string, "%3d(on)", volume - MINIMUM_LEVEL_2);
		break;
	default:
		sprintf(string, "%3d", volume - MINIMUM_LEVEL_2);
		break;
	}
}

/***************************************************************************
 *	deviceOldInfo: ġΥͥɽTOWNS Τͭ
 *	
 *		int deviceOldInfo(int mixer, const char *target, 
 *						int devices, int streo, int counter, int co);
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *target;		ܥ塼ǥХ̾
 *		int devices;			Ѳǽʥܥ塼Υӥåȥޥ
 *		int streo;				ƥ쥪ܥ塼Υӥåȥޥ
 *		int counter;			ܥ塼ǥХֹ
 *		int co;					ߤɽ
 *	֤
 *		ߤɽʾˣˤʤ
 *	
 *		
 ***************************************************************************/
int deviceOldInfo(int mixer, const char *target, 
						int devices, int streo, int counter, int co)
{
	int status;
	int  bitCounter;
	int level, left, right;
	int muteOnly = TOWNS_MUTE_DEVICES;
	char leftString[40];
	char rightString[40];

	bitCounter = (0x0001 << counter);

	if((OldDevicesInfo & devices) != 0 &&
			(target == NULL || name2target(target, bitCounter) == counter)) {
		switch(((devices & bitCounter) >> counter) +
				((streo & bitCounter) >> counter)) {
		default:
			break;
		case 1:	/* Mono */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			level &= 0x000000ff;
			if(muteOnly & bitCounter || level == 0) {
				oldLevel( leftString, USStoTowns66(level));
				printf("%6s  monoral %7s\n", 
						CurrentDevicesNames->names[counter], leftString);
			}
			break;
		case 2:	/* Stereo */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			left  = level & 0x000000ff;
			right = (level & 0x0000ff00) >> 8;
			oldLevel( leftString, USStoTowns66( left));
			oldLevel(rightString, USStoTowns66(right));
			printf("%6s  stereo %7s(L) %7s(R)\n", 
				CurrentDevicesNames->names[counter], leftString, rightString);
			break;
		}
	}
	return(0);
}
#endif /*TOWNS*/

/***************************************************************************
 *	levelGraph: ġΥǥХ򥰥ɽ
 *	
 *		int levelGraph(int level, int length);
 *	
 *		int level;			ܥ塼٥0100
 *		int length;			٥롦᡼Ĺ
 *	֤
 *		ɽʾ length ˤʤ
 *	
 *		
 ***************************************************************************/
int levelGraph(int level, int length)
{
	int counter;
	int mark[2] = {' ', '*'};

	for(counter = 1;counter <= length;++counter) {
		printf("%c", mark[(level > (100 * counter)/length)]);
	}
	return(length);
}

/***************************************************************************
 *	deviceGraph: ġΥǥХ򥰥ɽ
 *	
 *		int deviceGraph(int mixer, const char *target, 
 *						int devices, int streo, int counter, int co);
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *target;		ܥ塼ǥХ̾
 *		int devices;			Ѳǽʥܥ塼Υӥåȥޥ
 *		int streo;				ƥ쥪ܥ塼Υӥåȥޥ
 *		int counter;			ܥ塼ǥХֹ
 *		int co;					ߤɽ
 *	֤
 *		ߤɽʾˣˤʤ
 *	
 *		
 ***************************************************************************/
int deviceGraph(int mixer, const char *target, 
						int devices, int streo, int counter, int co)
{
	int status;
	int  bitCounter;
	int level, left, right;
#ifdef TOWNS
	int muteOnly = TOWNS_MUTE_DEVICES;
	const char *flag[] = {"off", "on"};
#endif /*TOWNS*/

	bitCounter = (0x0001 << counter);

	if(target == NULL || name2target(target, bitCounter) == counter) {
		switch(((devices & bitCounter) >> counter) +
				((streo & bitCounter) >> counter)) {
		default:
			break;
		case 1:	/* Mono */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			level &= 0x000000ff;
			printf("%6s  monoral   ", CurrentDevicesNames->names[counter]);
#ifdef TOWNS
			if(muteOnly & bitCounter) {
				printf(" %s\n", flag[level != 0]);
			} else {
				levelGraph(level, 10);
				printf("\n");
			}
#else
			levelGraph(level, 10);
			printf("\n");
#endif /*TOWNS*/
			break;
		case 2:	/* Stereo */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			left  = level & 0x000000ff;
			right = (level & 0x0000ff00) >> 8;
			printf("%6s  stereo  L:", CurrentDevicesNames->names[counter]);
			levelGraph( left, 10);
			printf(" R:");
			levelGraph(right, 10);
			printf("\n");
			break;
		}
	}
	return(0);
}

/***************************************************************************
 *	deviceUSS: ġΥǥХβɽ
 *	
 *		int deviceUSS(int mixer, const char *target,
 *						int devices, int streo, int counter, int co)
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *target;		ܥ塼ǥХ̾
 *		int devices;			Ѳǽʥܥ塼Υӥåȥޥ
 *		int streo;				ƥ쥪ܥ塼Υӥåȥޥ
 *		int counter;			ܥ塼ǥХֹ
 *		int co;					ߤɽ
 *	֤
 *		ߤɽʾˣˤʤ
 *	
 *		
 ***************************************************************************/
int deviceUSS(int mixer, const char *target, 
						int devices, int streo, int counter, int co)
{
	int status;
	int  bitCounter;
	int level, left, right;

	bitCounter = (0x0001 << counter);

	if(target == NULL || name2target(target, bitCounter) == counter) {
		switch(((devices & bitCounter) >> counter) +
				((streo & bitCounter) >> counter)) {
		default:
			break;
		case 1:	/* Mono */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			level &= 0x000000ff;
			printf("%6s  monoral Level%3d\n", 
					CurrentDevicesNames->names[counter], level);
			break;
		case 2:	/* Stereo */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			left  = level & 0x000000ff;
			right = (level & 0x0000ff00) >> 8;
			printf("%6s  stereo  Left %3d / Right %3d\n", 
					CurrentDevicesNames->names[counter], left, right);
			break;
		}
	}
	return(0);
}

/***************************************************************************
 *	deviceInfo: ġΥǥХβɽ
 *	
 *		int deviceInfo(int mixer, const char *target,
 *						int devices, int streo, int counter, int co)
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *target;		ܥ塼ǥХ̾
 *		int devices;			Ѳǽʥܥ塼Υӥåȥޥ
 *		int streo;				ƥ쥪ܥ塼Υӥåȥޥ
 *		int counter;			ܥ塼ǥХֹ
 *		int co;					ߤɽ
 *	֤
 *		ߤɽʾˣˤʤ
 *	
 *		
 ***************************************************************************/
int deviceInfo(int mixer, const char *target, 
						int devices, int streo, int counter, int co)
{
	int status;
	int bitCounter;
	int level, left, right;

	bitCounter = (0x0001 << counter);

	if(target == NULL || name2target(target, bitCounter) == counter) {
		switch(((devices & bitCounter) >> counter) +
				((streo & bitCounter) >> counter)) {
		default:
			break;
		case 1:	/* Mono */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			level &= 0x000000ff;
			printf("%6s  monoral Level%4.1f\n", 
					CurrentDevicesNames->names[counter], level / 10.0);
			break;
		case 2:	/* Stereo */
			status = ioctl(mixer, MIXER_READ(counter), &level);
			left  = level & 0x000000ff;
			right = (level & 0x0000ff00) >> 8;
			printf("%6s  stereo  Left %4.1f / Right %4.1f\n", 
						CurrentDevicesNames->names[counter], 
									left / 10.0, right / 10.0);
			break;
		}
	}
	return(0);
}

/***************************************************************************
 *	deviceList: ܥ塼ǥХΥꥹȤɽ
 *	
 *		int deviceList(int mixer, const char *target, int devMask, 
 *				int (*print)(int mixer, const char *target, 
 *							int devices, int streo, int counter, int co),
 *							int co);
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *target;		ܥ塼ǥХ̾
 *		int (*print)();			ġΥǥХɽؿ
 *		nt co;					ߤɽ
 *	֤
 *		ߤɽ
 *	
 *		
 ***************************************************************************/
int deviceList(int mixer, const char *target, int devMask, 
				int (*print)(int mixer, const char *target, 
						int devices, int streo, int counter, int co), int co)
{
	int devices;
	int streoDevices;
	int counter, bitCounter;

	if(ioctl(mixer, SOUND_MIXER_READ_DEVMASK, &devices) >= 0 &&
			ioctl(mixer, SOUND_MIXER_READ_STEREODEVS, &streoDevices) >= 0) {
		devices &= devMask;
		streoDevices &= devMask;
		for(counter = 0, bitCounter = 0x001;
				counter < SOUND_MIXER_NRDEVICES;
						++counter, bitCounter <<= 1) {
			if(print != NULL) {
				co = print(mixer, target, devices, streoDevices, counter, co);
			}
		}
	} else {
		printf("Unknown devices. \n");
	}
	return(co);
}

/***************************************************************************
 *	findInArray: ʸ󤫤θ
 *	
 *		int findInArray(const char *target, const char **array);
 *	
 *		const char *target;
 *		const char **array;
 *	֤
 *		EOF:	ꤵ줿ʸ󤬸Ĥʤä
 *		0ʾ:	Ĥäʸźֹ
 *	
 *		NULL ݥ󥿤Ǥʸ󤫤ꤵ줿ʸ򸡺롣
 ***************************************************************************/
int findInArray(const char *target, const char **array)
{
	int counter;

	for(counter = 0;*(array + counter) != NULL;++counter) {
		if(strcmp(*(array + counter), target) == 0) {
			return(counter);
		}
	}
	return(EOF);
}

/***************************************************************************
 *	string2level: ʸܥ塼٥Ѵ
 *	
 *		int string2level(const char *target, int ovol);
 *	
 *		const char *target;		̻ʸ(on/off/min/[dB/v/u])
 *		int ovol;				β̤
 *	֤
 *		ܥ塼٥0100
 *	
 *		
 ***************************************************************************/
int string2level(const char *target, int ovol)
{
	int level = EOF;
	char buffer[1024];
	double levelValue;
#ifdef	TOWNS
	int index;
	const char *levelString[] = {"off", "min", "on"};
	const int levelValues[] = 
					{MINIMUM_LEVEL_MUTE, MINIMUM_LEVEL_1, MAXIMUM_LEVEL};
#endif	/*TOWNS*/

	switch(*target) {
#ifdef	TOWNS
	case 'o':
	case 'm':
		index = findInArray(target, levelString);
		if(index != EOF && 
					index < (sizeof(levelValues)/sizeof(levelValues[0]))) {
			level = Towns66toUSS(levelValues[index]);
		}
		break;
#endif	/*TOWNS*/
	case '-':
		if(strncmp(target, "--", 2) == 0) {
			if(strcmp(target, "--") == 0) {
				level = ovol;
			}
			break;
		}
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		sscanf(target, "%lf%s", &levelValue, buffer);
		switch(buffer[0]) {
#ifdef	TOWNS
		case 'd':
			level = levelValue * 2 + MAXIMUM_LEVEL;
			level = Towns66toUSS(level);
			break;
#endif	/*TOWNS*/
		default:
#ifdef	TOWNS
			level = levelValue + 2;
			level = Towns66toUSS(level);
			break;
#endif	/*TOWNS*/
		case 'u':
			level = levelValue;
			break;
		case 'v':
			level = levelValue * 10;
			break;
		}
		break;
	default:
		break;
	}
	if(level > 100 || level < 0) {
		level = EOF;
	}
	return(level);
}

/***************************************************************************
 *	setVolume: ܥ塼٥ꤹ
 *	
 *		int setVolume(int mixer, int *argc_ptr, char ***argv_ptr);
 *	
 *		int mixer;			/dev/mixer Υե롦ǥץ
 *		int *argc_ptr;		θĿ򼨤ѿؤΥݥ
 *		char ***argv_ptr;	ʸ󷲤򼨤ؤΥݥ󥿤ؤΥݥ
 *		int (*print)(int mixer, const char *target, 
 *					int devices, int streo, int counter, int co);
 *	֤
 *		0:		˼
 *		0ʳ:	ꤷǥХΥӥåȥޥｪλ
 *	
 *		
 ***************************************************************************/
int setVolume(int mixer, int *argc_ptr, char ***argv_ptr,
			int (*print)(int mixer, const char *target, 
					int devices, int streo, int counter, int co))
{
	int deviceNo;
	int level, left, right, olevel;
	int devices;
	int streoDevices;
	int result = 0;
	int bitCounter;
	int argc = *argc_ptr;
	char **argv = *argv_ptr;
	const char *target;

	if(ioctl(mixer, SOUND_MIXER_READ_DEVMASK, &devices) < 0 ||
			ioctl(mixer, SOUND_MIXER_READ_STEREODEVS, &streoDevices) < 0) {
		fprintf(stderr, "Can't use /dev/mixer\n");
		return(result);
	}
	deviceNo = name2target(++*argv, devices);
	/*printf("Target:%s(%#x) #%d\n", *argv, devices, deviceNo);*/
	target = *argv;
	bitCounter = (0x0001 << deviceNo);
	if(--argc > 0 && deviceNo >= 0) {
		ioctl(mixer, MIXER_READ(deviceNo), &olevel);
		++argv;
		left = level = string2level((const char*)*argv, olevel & 0x000000ff);
		if(argc > 1) {
			right = string2level(*(argv+1), (olevel & 0x0000ff00) >> 8);
			if(right != EOF) {
				--argc, ++argv;
				level |= left | (right << 8);
			} else {
				level |= (level << 8);
			}
		} else {
			level |= (level << 8);
		}
		/*printf("Level:%s -> %d\n", *argv, level);*/
		if(level >= 0) {
			ioctl(mixer, MIXER_WRITE(deviceNo), &level);
			result |= bitCounter;
		}
	/*
	 *	if(print != NULL) {
	 *		(void)print(mixer, target, devices, streoDevices, deviceNo, 0);
	 *	}
	 */
	}
	*argc_ptr = argc;
	*argv_ptr = argv;
	return(result);
}

/***************************************************************************
 *	usage: ˡɽ
 *	
 *		void usage(int mixer, const char *program);
 *	
 *		int mixer;				/dev/mixer Υե롦ǥץ
 *		const char *program;	裰ʥץ̾
 *	
 *		
 ***************************************************************************/
void usage(int mixer, const char *program)
{
	int streoDevices;

	if(ioctl(mixer, SOUND_MIXER_READ_STEREODEVS, &streoDevices) >= 0) {
		fprintf(stderr,"Usage:\n");
#ifdef	TOWNS
		fprintf(stderr,"       %s [-[Hhl[gauvto]UTL]] [device...] \n", program);
		fprintf(stderr,"       %s [-v[gauvto]UTL] --volume-device vol\n", program);
#else
		fprintf(stderr,"       %s [-[Hhl[gauv]L]] [device...] \n", program);
		fprintf(stderr,"       %s [-v[gauvto]L] --volume-device vol\n", program);
#endif	/*TOWNS*/
#ifdef	TOWNS
		fprintf(stderr,"       %s chan vol\n", program);
		fprintf(stderr,"\n");
		fprintf(stderr,"   chan: ll .... line in (L)\n");
		fprintf(stderr,"         lr .... line in (R)\n");
		if(streoDevices & SOUND_MASK_ALTPCM) {
			fprintf(stderr,"         al .... AD/DA in (L)\n");
			fprintf(stderr,"         al .... AD/DA in (L)\n");
		}
		fprintf(stderr,"         cr .... CD-DA in (R)\n");
		fprintf(stderr,"         cr .... CD-DA in (R)\n");
		fprintf(stderr,"         mc .... microphone\n");
		fprintf(stderr,"    vol: off .... turn off\n");
		fprintf(stderr,"         on  .... 0db\n");
		fprintf(stderr,"         min .... -32.0db\n");
		fprintf(stderr,"         0   .... -31.5db\n");
		fprintf(stderr,"         63  .... 0db\n");
		fprintf(stderr,"         -32.0db -- 0db\n");
#else
		fprintf(stderr,"    vol:\n");
#endif	/*TOWNS*/
		fprintf(stderr,"         0u -- 100u    \n");
		fprintf(stderr,"         0v -- 10.0v   \n");
	} else {
		fprintf(stderr, "Can't use /dev/mixer.\n");
	}
}

/***************************************************************************
 *	main: ᥤؿ
 *	
 *		int main(int argc, char **argv);
 *	
 *		int argc;		θĿ
 *		char **argv;	ʸ
 *	֤
 *		0:		ｪλ
 *		0ʳ:	顼
 *	
 *		
 ***************************************************************************/
int main(int argc, char **argv)
{
	int mixer;
	int co = 0;
	char *columns;
	char *prog = *argv;
	int loopOut = 0;
	int c = 0, options = 0;
	int currentDeviceName = 0;
	int adjustedDevicesBits = 0;
	int devMask = (0x0001 << SOUND_MIXER_NRDEVICES) - 1;
	DeviceInfoProc def_device_verbose = deviceInfo;
	DeviceInfoProc device_verbose = NULL;
	DeviceInfoProc def_device_print = devicePrint;
	DeviceInfoProc device_print = NULL;
	DeviceInfoProc current_device_disp = NULL;
#ifdef	TOWNS
	int status;
#endif	/*TOWNS*/

	mixer = open("/dev/mixer", O_RDONLY);
	columns = getenv("COLUMNS");
	if(columns != NULL) {
		sscanf(columns, "%d", &Columns);
	}
	if(mixer >= 0) {
		while (--argc > 0) {
			if(**++argv == '-') {
				loopOut = 0;
				while ((c = *++*argv) != 0) {
					switch (c) {
					case '-':
						if(device_verbose != NULL && current_device_disp != NULL) {
							device_verbose = current_device_disp;
						}
						adjustedDevicesBits |= 
							setVolume(mixer, &argc, &argv, device_verbose);
						++options;
						loopOut = 1;
						break;
					case 'h':
					case 'H':
						++options;
						usage(mixer, prog);
						break;
					case 'g':
						current_device_disp = deviceGraph;
						break;
					case 'u':
						current_device_disp = deviceUSS;
						break;
					case 'l':
						if(device_print == NULL) {
							device_print = deviceInfo;
						}
						break;
					case 'a':
						current_device_disp = deviceInfo;
						break;
					case 'v':
						if(device_verbose == NULL) {
							device_verbose = def_device_verbose;
						}
						break;
					case 'L':
						currentDeviceName |= 0x01;
						CurrentDevicesNames = &DevicesNames[currentDeviceName];
						break;
					case 'V':
						printf("%s\n", EVOL_VERSION);
						exit(0);
						break;
#ifdef	TOWNS
					case 'U':
						currentDeviceName |= 0x02;
						CurrentDevicesNames = &DevicesNames[currentDeviceName];
						break;
					case 'T':
						currentDeviceName &= 0x01;
						CurrentDevicesNames = DevicesNames;
						break;
					case 't':
						current_device_disp = device_verbose = deviceInfoTowns;
						break;
					case 'o':
						current_device_disp = device_verbose = deviceOldInfo;
						break;
#endif	/*TOWNS*/
					default:
						++options;
						fprintf(stderr, "%s: illegal option %c\n", prog, c);
						argc = 0;
						usage(mixer, prog);
						exit(0);
						break;
					}
					if(loopOut) {
						break;
					}
				}
#ifdef	TOWNS
			} else {
				status = townsStandardTarget(mixer, &argc, &argv, device_verbose);
				if(status != 0) {
					adjustedDevicesBits |= status;
					++options;
				} else {
					break;
				}
#endif	/*TOWNS*/
			}
		}
		if(device_print == NULL) {
			if(current_device_disp == NULL) {
				device_print = def_device_print;
			}
		} else {
			if(current_device_disp != NULL) {
				device_print = current_device_disp;
			}
		}
		if(device_verbose != NULL) {
			co = deviceList(mixer, NULL, adjustedDevicesBits, device_verbose, co);
			devMask = ~adjustedDevicesBits;
		}
		if(device_print != NULL) {
			if(options == 0) {		/* ʤ */
				if(argc == 0) {		/* ʤ */
					co = deviceList(mixer, NULL, devMask, device_print, co);
					devMask = 0x0000;
				}
			}
			while(--argc >= 0) {
				co = deviceList(mixer, *argv, devMask, device_print, co);
				++argv;
			}
		} else {
			if(options == 0 && device_print == NULL) {
				usage(mixer, prog);
			}
		}
		if(ferror(stdout)) {
			fprintf(stderr, "%s: error writting stdout\n", prog);
			exit(2);
		}
		if(co > 0) {
			printf("\n");
		}
	} else {
		printf("Can't Open.\n");
	}
	close(mixer);
	return(0);
}

